package database.generic;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.TransactionManager;
import javax.transaction.UserTransaction;


import org.enhydra.jdbc.pool.StandardXAPoolDataSource;
import org.enhydra.jdbc.standard.StandardXADataSource;
import org.objectweb.transaction.jta.TMService;

/**
 * Base class for all tests to test the database functionality.
 */
public class DatabaseTest
{
    // Default database properties if no others are supplied ////////////////////

    /**
     * Default name for database user.
     */
    protected static String DEFAULT_DB_USER = "mojo";

    /**
     * Default password for database user.
     */
    protected static String DEFAULT_DB_PASSWORD = "jojo";

    /**
     * Default initial database connection count in connection pool.
     */
    protected static String DEFAULT_INIT_CONNECTION_COUNT = "3";

    /**
     * Default minimal database connection count in connection pool.
     */
    protected static String DEFAULT_MIN_CONNECTION_COUNT = "1";

    /**
     * Default maximal database connection count in connection pool.
     */
    protected static String DEFAULT_MAX_CONNECTION_COUNT = "5";

    /**
     * Default user name to use if no other is specified.
     */
    protected static String m_strDefaultUserName = "mojo";

    /**
     * Default pasword to use if no other is specified.
     */
    protected static String m_strDefaultPassword = "jojo";

    // Transaction Manager shared between all tests /////////////////////////////

    /**
     * Instance of JOTM transaction manager.
     */
    protected static TMService s_tmService;

    /**
     * Instance of transaction manager we are using.
     */
    protected static TransactionManager s_transactionManager;

    /**
     * Registered data sources.
     */
    protected static Hashtable s_htblDataSources;

    // Database connection parameters for a test ////////////////////////////////

    /**
     * Data source used for all test.
     */
    protected StandardXAPoolDataSource m_pooledDataSource;

    /**
     * Database transaction which can be used by test.
     */
    protected UserTransaction m_transaction;

    /**
     * Connection to the database which can be used by test.
     */
    protected Connection m_connection;

    /**
     * Static initializer
     */
    static
    {
        s_htblDataSources = new Hashtable();

        System.out.println("Starting JoTM transaction manager.");
        try
        {
            // Start the transaction manager
            s_tmService = new org.objectweb.jotm.Jotm(true, true);

            //  Register the user transaction object in JNDI
            // TODO: This needs to be redone, email from Jeff:
            // For example, a J2EE app server requires the UserTransaction
            // to be available through JNDI with the name
            // "java:comp/env/UserTransaction" (ENC requirements).
            // What's more, using resource factories, you
            // don't even bind UserTransaction object
            // (you create instead an object inheriting from
            // javax.naming/spi/ObjectFactory which'll return an UserTransaction object).
            // On the other hand, if an application is not supporting
            // ENC, UserTransaction can be available with
            // the name "javax.transaction.UserTransaction" (see the JDBC example of JOTM)
            // I prefer not to mandate a specific use of JOTM so that it can fit into several use cases.

            Context ictx = null;
            String  userTransactionName = "javax.transaction.UserTransaction";
            String  transactionManagerName = "javax.transaction.TransactionManager";

            ictx = new InitialContext();

            if (userTransactionName != null)
            {
                ictx.rebind(userTransactionName, s_tmService.getUserTransaction());
                System.out.println("UserTransaction object bound in JNDI with name "
                        + userTransactionName);
            }

            if (transactionManagerName != null)
            {
                ictx.rebind(transactionManagerName, s_tmService.getTransactionManager());
                System.out.println("TransactionManager object bound in JNDI with name "
                        + transactionManagerName);
            }
            System.out.println("JoTM is running...");

            // Construct transaction manager
            s_transactionManager = s_tmService.getTransactionManager();
        }
        catch (NamingException namingExc)
        {
            throw new RuntimeException(namingExc);
        }
    }

    /**
     * Create new DatabaseTest.
     *
     * @param strTestName - name of the test
     * @param strDataSourceName - data source which will be used by this test
     * @throws NumberFormatException
     * @throws Exception
     */
    public DatabaseTest(
            String strTestName,
            String strDataSourceName
            ) throws NumberFormatException,
            Exception
    {


        m_pooledDataSource = (StandardXAPoolDataSource)s_htblDataSources.get(
                strDataSourceName);
        if (m_pooledDataSource == null)
        {
            System.err.println("The datasource with name " + strDataSourceName
                    + " doesn't exist");
        }

        m_transaction = null;
        m_connection = null;
    }

    /**
     * Create data source to be managed by the transaction manager.
     * Each test must specify the name of the data source. If it doesn't exist,
     * it will be created.
     *
     * @param strDataSourceName - data source which will be used by this test
     * @param strDriverName - driver name which will be used by this test
     * @param strUrl - url this test connects to the database
     */
    public static synchronized void addDataSource(
            String  strDataSourceName,
            String  strDriverName,
            String  strUrl
            ) throws NumberFormatException, Exception
    {
        addDataSource(strDataSourceName, strDriverName, strUrl,
                DEFAULT_DB_USER, DEFAULT_DB_PASSWORD,
                DEFAULT_INIT_CONNECTION_COUNT,
                DEFAULT_MIN_CONNECTION_COUNT,
                DEFAULT_MAX_CONNECTION_COUNT);
    }

    /**
     * Create data source to be managed by the transaction manager.
     * Each test must specify the name of the data source. If it doesn't exist,
     * it will be created.
     *
     * @param strDataSourceName - data source which will be used by this test
     * @param strDriverName - driver name which will be used by this test
     * @param strUrl - url this test connects to the database
     * @param strUser - user name to connects to the database
     * @param strPassword - password to connects to the database
     * @param strInitialConnectionCount - initial connection count
     * @param strMinimalPoolSize - minimal connection count
     * @param strMaximalPoolSize - maximal connection count
     */
    public static synchronized void addDataSource(
            String  strDataSourceName,
            String  strDriverName,
            String  strUrl,
            String  strUser,
            String  strPassword,
            String  strInitialConnectionCount,
            String  strMinimalPoolSize,
            String  strMaximalPoolSize
            ) throws NumberFormatException, Exception
    {
        if (s_htblDataSources.get(strDataSourceName) == null)
        {
            // create an XA pool datasource with a minimum of 4 objects
            StandardXAPoolDataSource xapdsPoolDataSource = new StandardXAPoolDataSource(
                    Integer.parseInt(strInitialConnectionCount));
            xapdsPoolDataSource.setMinSize(Integer.parseInt(strMinimalPoolSize));
            xapdsPoolDataSource.setMaxSize(Integer.parseInt(strMaximalPoolSize));
            xapdsPoolDataSource.setUser(strUser);
            xapdsPoolDataSource.setPassword(strPassword);

            StandardXADataSource xadsDataSource;

            xadsDataSource = new StandardXADataSource();
            xadsDataSource.setDriverName(strDriverName);
            xadsDataSource.setUrl(strUrl);
            xadsDataSource.setUser(strUser);
            xadsDataSource.setPassword(strPassword);
            xadsDataSource.setTransactionManager(s_transactionManager);

            // Transaction manager was already created so attach it with DS
            xapdsPoolDataSource.setDataSourceName("XA" + strDataSourceName);
            xapdsPoolDataSource.setDataSource(xadsDataSource);
xapdsPoolDataSource.setTransactionManager(s_transactionManager);
            s_htblDataSources.put(strDataSourceName, xapdsPoolDataSource);

            System.out.println("DataSource " + strDataSourceName + " registered.");
        }
    }

    /**
     * Retrieve connection for default user and data source.
     *
     * @return Connection - database connection for default user and data source.
     * @throws SQLException
     * @throws SystemException
     * @throws RollbackException
     */
    public Connection getConnection(
            ) throws SQLException,
            SystemException,
            RollbackException
    {
        return m_pooledDataSource.getConnection();
    }

    /**
     * Retrieve connection for default user and data source.
     *
     * @param strUserName
     * @param strPassword
     * @return Connection - database connection for default user and data source.
     * @throws SQLException
     * @throws SystemException
     * @throws RollbackException
     */
    public Connection getConnection(
            String strUserName,
            String strPassword
            ) throws SQLException,
            SystemException,
            RollbackException
    {
        return m_pooledDataSource.getConnection(strUserName, strPassword);
    }

    /*
    * @see TestCase#setUp()
    */
    protected void setUp() throws Exception
    {

        // Get connection
        System.out.println("GET THE CONNECTION");
        m_connection = getConnection();
        m_transaction = s_tmService.getUserTransaction();
    }

    /*
    * @see TestCase#tearDown()
    */
    protected void tearDown() throws Exception
    {
        try
        {
            if (m_transaction.getStatus() != Status.STATUS_NO_TRANSACTION)
            {
                System.err.println("Transaction wasnt commited or rollbacked.");
            }
        }
        finally
        {
            System.out.println("-----------1");
            if (m_connection != null)
            {
                            System.out.println("-----------2");
                if (!m_connection.isClosed())
                {
                                System.out.println("-----------3");
                    m_connection.close();
                                System.out.println("-----------4");
                }
            }
        }

    }

    /**
     * Execute given statement and close it after it is done.
     *
     * @param jdbcstatement - statement to execute
     * @param strStatement - textual representation of statement to print
     * @return int - number of affected rows
     * @throws SQLException
     */
    protected static int executeUpdate(
            PreparedStatement jdbcstatement,
            String            strStatement
            ) throws SQLException
    {
        int iUpdateCount = 0;

        try
        {
            System.out.println("Executing statement " + strStatement);
            iUpdateCount = jdbcstatement.executeUpdate();
            System.out.println("Statement executed....");
        }
        finally
        {
            jdbcstatement.close();
        }

        return iUpdateCount;
    }

    /**
     * Gracefully close result set and statement so that no error is generated.
     *
     * @param results - result set to close, if null it is ignored
     * @param jdbcstatement - jdbc statement to close, if null it is ignored
     */
    protected static void closeResultSetAndStatement(
            ResultSet results,
            Statement jdbcstatement
            ) throws SQLException
    {
        try
        {
            if (results != null)
            {
                results.close();
            }
        }
        catch(SQLException sqlExc)
        {
            // TODO: Log it here somehow and throw it if we want to
            throw sqlExc;
        }
        finally
        {
            if (jdbcstatement != null)
            {
                try
                {
                    jdbcstatement.close();
                }
                catch(SQLException sqlExc)
                {
                    // TODO: Log it here somehow and throw it if we want to
                    throw sqlExc;
                }
            }
        }
    }
}
